# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
from unittest import TestCase

from hamcrest import assert_that, equal_to, only_contains, raises

from storops.exception import UnityException, UnitySmbShareNameExistedError, \
    UnityAclUserNotFoundError, UnitySnapNameInUseError
from storops.unity.enums import CIFSTypeEnum, ACEAccessTypeEnum, \
    CifsShareOfflineAvailabilityEnum

from storops.unity.resource.cifs_share import UnityCifsShare, \
    UnityCifsShareList, UnityAclUser, UnityAclUserList
from storops_test.unity.rest_mock import t_rest, patch_rest

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)

snap_based_share = UnityCifsShare(cli=t_rest(), _id='SMBShare_5')


class UnityCifsShareTest(TestCase):
    @patch_rest
    def test_properties(self):
        cifs = UnityCifsShare('SMBShare_1', cli=t_rest())
        assert_that(cifs.id, equal_to('SMBShare_1'))
        assert_that(cifs.type, equal_to(CIFSTypeEnum.CIFS_SHARE))
        assert_that(cifs.offline_availability, equal_to(
            CifsShareOfflineAvailabilityEnum.MANUAL))
        assert_that(cifs.name, equal_to('esa_cifs1'))
        assert_that(cifs.path, equal_to(r'/'))
        assert_that(cifs.export_paths, only_contains(
            r'\\smb1130.win2012.dev\esa_cifs1',
            r'\\10.244.220.120\esa_cifs1'))
        assert_that(cifs.description, equal_to('abc'))
        assert_that(str(cifs.creation_time),
                    equal_to("2016-03-02 02:43:34.014000+00:00"))
        assert_that(str(cifs.modified_time),
                    equal_to("2016-03-02 02:43:34.014000+00:00"))
        assert_that(cifs.is_continuous_availability_enabled, equal_to(False))
        assert_that(cifs.is_encryption_enabled, equal_to(False))
        assert_that(cifs.is_ace_enabled, equal_to(False))
        assert_that(cifs.is_abe_enabled, equal_to(False))
        assert_that(cifs.is_branch_cache_enabled, equal_to(False))
        assert_that(cifs.is_dfs_enabled, equal_to(False))
        assert_that(cifs.umask, equal_to('022'))
        assert_that(cifs.filesystem.get_id(), equal_to('fs_2'))

    @patch_rest
    def test_cifs_server(self):
        share = UnityCifsShare('SMBShare_1', cli=t_rest())
        assert_that(share.cifs_server.name, equal_to('nas1130'))
        assert_that(share.cifs_server.domain, equal_to('win2012.dev'))

    @patch_rest
    def test_get_all(self):
        cifs_list = UnityCifsShareList(cli=t_rest())
        assert_that(len(cifs_list), equal_to(1))

    @patch_rest
    def test_create_path_not_exists(self):
        def f():
            UnityCifsShare.create(t_rest(), 'cs1', 'fs_8', '/cs1')

        assert_that(f, raises(UnityException,
                              'could not find the specified path'))

    @patch_rest
    def test_create_success(self):
        share = UnityCifsShare.create(t_rest(), 'cs1', 'fs_8')
        assert_that(share.name, equal_to('cs1'))

    @patch_rest
    def test_create_same_name_exists(self):
        def f():
            UnityCifsShare.create(t_rest(), 'cs2', 'fs_8')

        assert_that(f, raises(UnitySmbShareNameExistedError, 'already exists'))

    @patch_rest
    def test_delete_share_success(self):
        share = UnityCifsShare(_id='SMBShare_7', cli=t_rest())
        resp = share.delete()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_get_by_name(self):
        shares = UnityCifsShareList(cli=t_rest(), name='cs1')
        assert_that(len(shares), equal_to(1))
        share = shares[0]
        assert_that(share.name, equal_to('cs1'))

    @patch_rest
    def test_create_snap_existed(self):
        def f():
            share = UnityCifsShare(cli=t_rest(), _id='SMBShare_8')
            share.create_snap('share_snap')

        assert_that(f, raises(UnitySnapNameInUseError, 'in use'))

    @patch_rest
    def test_delete_snap_based_share(self):
        share = UnityCifsShare(cli=t_rest(), _id='SMBShare_15')
        resp = share.delete()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_get_domain_user_name_default_domain(self):
        share = UnityCifsShare(cli=t_rest(), _id='SMBShare_8')
        name = share._get_domain_user_name(user='admin')
        assert_that(name, equal_to(r'win2012.dev\admin'))

    @patch_rest
    def test_add_ace_success(self):
        share = snap_based_share
        resp = share.add_ace(user='administrator')
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_add_ace_no_user(self):
        def f():
            share = UnityCifsShare(cli=t_rest(), _id='SMBShare_5')
            share.add_ace()

        assert_that(f, raises(ValueError, 'username'))

    @patch_rest
    def test_enabled_ace_success(self):
        share = UnityCifsShare(cli=t_rest(), _id='SMBShare_8')
        resp = share.enable_ace()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_disable_ace_success(self):
        share = UnityCifsShare(cli=t_rest(), _id='SMBShare_8')
        resp = share.disable_ace()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_ace_list_success(self):
        share = UnityCifsShare(cli=t_rest(), _id='SMBShare_5')
        ace_list = share.get_ace_list()
        assert_that(len(ace_list), equal_to(2))

        ace = ace_list[0]
        assert_that(ace.access_type, equal_to(ACEAccessTypeEnum.GRANT))

    @patch_rest
    def test_ace_clear_access_success(self):
        share = UnityCifsShare(cli=t_rest(), _id='SMBShare_5')
        removed_sid_list = share.clear_access()
        assert_that(removed_sid_list,
                    only_contains('S-1-5-15-be80fa7-8ddad211-d49ba5f9-45e',
                                  'S-1-5-15-be80fa7-8ddad211-d49ba5f9-1f4'))

    @patch_rest
    def test_ace_clear_access_with_white_list(self):
        share = UnityCifsShare(cli=t_rest(), _id='SMBShare_5')
        removed_sid_list = share.clear_access(white_list=['SMIS_User_2'])
        assert_that(removed_sid_list,
                    equal_to(['S-1-5-15-be80fa7-8ddad211-d49ba5f9-45e']))

    @patch_rest
    def test_delete_ace(self):
        share = UnityCifsShare(cli=t_rest(), _id='SMBShare_5')
        resp = share.delete_ace('win2012.dev', 'SMIS_User_2')
        assert_that(resp.is_ok(), equal_to(True))


class UnityAclUserTest(TestCase):
    @patch_rest
    def test_properties(self):
        _id = 'S-1-5-15-be80fa7-8ddad211-d49ba5f9-452'
        user = UnityAclUser(cli=t_rest(), _id=_id)
        assert_that(user.existed, equal_to(True))
        assert_that(user.id, equal_to(_id))
        assert_that(user.sid, equal_to(_id))
        assert_that(user.user_name, equal_to('L1PFC239208$'))
        assert_that(user.domain_name, equal_to('win2012.dev'))

    @patch_rest
    def test_get_all(self):
        users = UnityAclUserList(cli=t_rest())
        assert_that(len(users), equal_to(3))

    @patch_rest
    def test_get_sid_found(self):
        sid = UnityAclUser.get_sid(t_rest(), 'administrator', 'win2012.dev')
        assert_that(sid, equal_to('S-1-5-15-be80fa7-8ddad211-d49ba5f9-1f4'))

    @patch_rest
    def test_get_sid_not_found(self):
        def f():
            UnityAclUser.get_sid(t_rest(), 'not_exists', 'win2012.dev')

        assert_that(f, raises(UnityAclUserNotFoundError))

    @patch_rest
    def test_sid_list(self):
        share = UnityCifsShare(cli=t_rest(), _id='SMBShare_5')
        ace_list = share.get_ace_list()
        assert_that(ace_list.sid_list, only_contains(
            'S-1-5-15-be80fa7-8ddad211-d49ba5f9-45e',
            'S-1-5-15-be80fa7-8ddad211-d49ba5f9-1f4'))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

import ddt
import mock
from hamcrest import assert_that, calling, only_contains, instance_of, \
    contains_string, raises, none, has_item, is_not
from hamcrest import equal_to
from storops.unity.resource.snap_schedule import UnitySnapSchedule

from storops import UnitySystem, TieringPolicyEnum
from storops.exception import UnitySnapNameInUseError, \
    UnityLunNameInUseError, UnityLunShrinkNotSupportedError, \
    UnityNothingToModifyError, UnityPerfMonNotEnabledError, \
    UnityThinCloneLimitExceededError, UnityCGMemberActionNotSupportError, \
    UnityThinCloneNotAllowedError, UnityMigrationTimeoutException, \
    UnityMigrationSourceDestNotExistsError, JobStateError, \
    JobTimeoutException
from storops.unity.enums import HostLUNAccessEnum, NodeEnum, RaidTypeEnum, \
    ESXFilesystemBlockSizeEnum, ESXFilesystemMajorVersionEnum
from storops.unity.resource.disk import UnityDisk
from storops.unity.resource.host import UnityBlockHostAccessList, UnityHost
from storops.unity.resource.lun import UnityLun, UnityLunList
from storops.unity.resource.pool import UnityPool
from storops.unity.resource.port import UnityIoLimitPolicy, \
    UnityIoLimitRuleSetting
from storops.unity.resource.remote_system import UnityRemoteSystem
from storops.unity.resource.snap import UnitySnap
from storops.unity.resource.sp import UnityStorageProcessor
from storops.unity.resource.storage_resource import UnityStorageResource
from storops.unity.resp import RestResponse
from storops_test.unity.jh_mock import MockJobHelper
from storops_test.unity.rest_mock import t_rest, t_unity, patch_rest
from storops_test.utils import is_nan

__author__ = 'Cedric Zhuang'


@ddt.ddt
class UnityLunTest(TestCase):
    @patch_rest
    def test_get_lun_sv2_simple_property(self):
        lun = UnityLun(_id='sv_2', cli=t_rest())
        assert_that(lun.existed, equal_to(True))
        assert_that(lun.id, equal_to('sv_2'))
        assert_that(lun.name, equal_to('openstack_lun'))
        assert_that(lun.description, equal_to('sample'))
        assert_that(lun.size_total, equal_to(107374182400))
        assert_that(lun.total_size_gb, equal_to(100))
        assert_that(lun.size_allocated, equal_to(0))
        assert_that(lun.per_tier_size_used, only_contains(2952790016, 0, 0))
        assert_that(lun.is_thin_enabled, equal_to(True))
        assert_that(lun.wwn, equal_to(
            '60:06:01:60:17:50:3C:00:C2:0A:D5:56:92:D1:BA:12'))
        assert_that(lun.is_replication_destination, equal_to(False))
        assert_that(lun.is_snap_schedule_paused, equal_to(False))
        assert_that(lun.metadata_size, equal_to(5100273664))
        assert_that(lun.metadata_size_allocated, equal_to(2684354560))
        assert_that(lun.snap_wwn, equal_to(
            '60:06:01:60:17:50:3C:00:C4:0A:D5:56:00:95:DE:11'))
        assert_that(lun.snaps_size, equal_to(0))
        assert_that(lun.snaps_size_allocated, equal_to(0))
        assert_that(lun.snap_count, equal_to(0))
        assert_that(lun.storage_resource, instance_of(UnityStorageResource))
        assert_that(lun.pool, instance_of(UnityPool))
        assert_that(lun.io_limit_rule, none())
        assert_that(lun.is_compression_enabled, equal_to(False))
        assert_that(lun.is_data_reduction_enabled, equal_to(False))

    @patch_rest
    def test_lun_modify_host_access(self):
        host = UnityHost(_id="Host_1", cli=t_rest())
        lun = UnityLun(_id='sv_4', cli=t_rest())
        host_access = [{'host': host, 'accessMask': HostLUNAccessEnum.BOTH}]
        lun.modify(host_access=host_access)
        lun.update()
        assert_that(lun.host_access[0].host, equal_to(host))
        assert_that(lun.host_access[0].access_mask,
                    equal_to(HostLUNAccessEnum.BOTH))

    @patch_rest
    def test_lun_modify_sp(self):
        lun = UnityLun(_id='sv_4', cli=t_rest())
        sp = UnityStorageProcessor(_id='spb', cli=t_rest())
        lun.modify(sp=sp)
        lun.update()
        assert_that(sp.to_node_enum(), equal_to(NodeEnum.SPB))

    @patch_rest
    def test_lun_modify_sp_with_id(self):
        lun = UnityLun(_id='sv_4', cli=t_rest())
        resp = lun.modify(sp=1)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_lun_modify_sp_with_enum(self):
        lun = UnityLun(_id='sv_4', cli=t_rest())
        resp = lun.modify(sp=NodeEnum.SPB)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_lun_modify_none(self):
        lun = UnityLun(_id='sv_4', cli=t_rest())
        resp = lun.modify(host_access=None)
        lun.update()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_lun_modify_wipe_host_access(self):
        lun = UnityLun(_id='sv_4', cli=t_rest())
        resp = lun.modify(host_access=[])
        lun.update()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_lun_modify_muitl_property_except_sp(self):
        lun = UnityLun(_id='sv_4', cli=t_rest())
        lun.modify(name="RestLun100", is_compression=True,
                   description="Lun description")
        lun.update()
        assert_that(lun.name, equal_to('RestLun100'))
        assert_that(lun.description, equal_to('Lun description'))

    @patch_rest
    def test_lun_modify_compression_enabled_v4_2(self):
        lun = UnityLun(_id='sv_17', cli=t_rest(version='4.2'))
        lun.modify(is_compression=True)
        lun.update()
        assert_that(lun.is_compression_enabled, equal_to(True))

    @patch_rest
    def test_lun_modify_compression_enabled(self):
        lun = UnityLun(_id='sv_18', cli=t_rest(version='4.3'))
        lun.modify(is_compression=True)
        lun.update()
        assert_that(lun.is_data_reduction_enabled, equal_to(True))

    @mock.patch(target='storops.lib.job_helper.JobHelper',
                new=MockJobHelper)
    @patch_rest
    def test_lun_delete_async(self):
        lun = UnityLun(_id='sv_1535', cli=t_rest())
        resp = lun.delete(force_snap_delete=True, force_vvol_delete=True)
        lun.update()
        assert_that(resp.is_ok(), equal_to(True))
        assert_that(resp.job.existed, equal_to(True))
        assert_that(resp.job.state.index, equal_to(4))

    @patch_rest
    def test_lun_delete_sync(self):
        lun = UnityLun(_id='sv_4', cli=t_rest())
        resp = lun.delete(async_mode=False, force_snap_delete=True,
                          force_vvol_delete=True)
        lun.update()
        assert_that(resp.is_ok(), equal_to(True))
        assert_that(resp.job.existed, equal_to(False))

    @patch_rest
    def test_lun_delete_thinclone(self):
        lun = UnityLun(_id='sv_5604', cli=t_rest())
        resp = lun.delete(async_mode=False, force_snap_delete=True,
                          force_vvol_delete=True)
        lun.update()
        assert_that(resp.is_ok(), equal_to(True))
        assert_that(resp.job.existed, equal_to(False))

    @patch_rest
    def test_lun_delete_has_thinclone(self):
        lun = UnityLun(_id='sv_5605', cli=t_rest())
        resp = lun.delete(force_snap_delete=True, force_vvol_delete=True)
        lun.update()
        assert_that(resp.is_ok(), equal_to(True))
        assert_that(resp.job.existed, equal_to(False))

    @mock.patch(target='storops.lib.job_helper.JobHelper',
                new=MockJobHelper)
    @patch_rest
    def test_lun_delete_async_job_state_error(self):
        def f():
            lun = UnityLun(_id='sv_1536', cli=t_rest())
            lun.delete(force_snap_delete=True, force_vvol_delete=True)

        assert_that(f, raises(JobStateError))

    @mock.patch(target='storops.lib.job_helper.JobHelper',
                new=MockJobHelper)
    @patch_rest
    def test_lun_delete_async_job_timeout(self):
        def f():
            lun = UnityLun(_id='sv_1537', cli=t_rest())
            lun.delete(force_snap_delete=True, force_vvol_delete=True,
                       async_timeout=3, async_interval=1)

        assert_that(f, raises(JobTimeoutException))

    @patch_rest
    def test_lun_attach_to_new_host(self):
        host = UnityHost(_id="Host_10", cli=t_rest())
        lun = UnityLun(_id='sv_4', cli=t_rest())
        resp = lun.attach_to(host)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_lun_attach_to_same_host(self):
        host = UnityHost(_id="Host_1", cli=t_rest())
        lun = UnityLun(_id='sv_4', cli=t_rest())
        resp = lun.attach_to(host, access_mask=HostLUNAccessEnum.BOTH)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_lun_attach_to_with_hlu(self):
        host = UnityHost(_id="Host_1", cli=t_rest())
        lun = UnityLun(_id='sv_6', cli=t_rest(version='4.4.0'))
        resp = lun.attach_to(host, access_mask=HostLUNAccessEnum.BOTH,
                             hlu=1655)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_lun_attach_to_without_hlu(self):
        host = UnityHost(_id="Host_1", cli=t_rest())
        lun = UnityLun(_id='sv_7', cli=t_rest(version='4.4.0'))
        resp = lun.attach_to(host, access_mask=HostLUNAccessEnum.BOTH)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_lun_detach_from_host(self):
        host = UnityHost(_id="Host_1", cli=t_rest())
        lun = UnityLun(_id='sv_16', cli=t_rest())
        resp = lun.detach_from(host)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_lun_detach_from_all_hosts(self):
        lun = UnityLun(_id='sv_5', cli=t_rest())
        lun.is_cg_member = False
        resp = lun.detach_from(host=None)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_get_lun_sv2_nested_property_update_property(self):
        lun = UnityLun(_id='sv_2', cli=t_rest())
        sr = lun.storage_resource
        assert_that(sr._cli, equal_to(t_rest()))
        assert_that(sr.size_total, equal_to(107374182400))

    @patch_rest
    def test_get_lun_sv3_nested_property_no_update(self):
        lun = UnityLunList.get(_id='sv_3', cli=t_rest())
        sr = lun.storage_resource
        assert_that(sr._cli, equal_to(t_rest()))

    @patch_rest
    def test_get_lun_all_0(self):
        lun_list = UnityLunList.get(cli=t_rest())
        assert_that(len(lun_list), equal_to(5))

    @patch_rest
    def test_get_lun_doc(self):
        lun = UnityLun(_id='sv_2', cli=t_rest())
        doc = lun.doc
        assert_that(doc,
                    contains_string('Represents Volume, LUN, Virtual Disk.'))
        assert_that(doc, contains_string('current_node'))
        assert_that(doc, contains_string('Current SP'))

    @patch_rest
    def test_get_lun_with_host_access(self):
        unity = UnitySystem('10.109.22.101', 'admin', 'Password123!')
        lun = unity.get_lun(_id='sv_567')
        assert_that(lun.host_access, instance_of(UnityBlockHostAccessList))
        access = lun.host_access[0]
        assert_that(access.access_mask, equal_to(HostLUNAccessEnum.PRODUCTION))
        assert_that(access.host, instance_of(UnityHost))
        assert_that(access.host.id, equal_to('Host_1'))
        assert_that(lun.host_access.get_host_id(), equal_to([access.host.id]))

    @patch_rest
    def test_lun_snap_create(self):
        lun = UnityLun(_id='sv_8', cli=t_rest())
        snap = lun.create_snap(name='lun_snap_1')
        assert_that(snap, instance_of(UnitySnap))

    @patch_rest
    def test_lun_snapshots(self):
        lun = UnityLun(_id='sv_8', cli=t_rest())
        assert_that(len(lun.snapshots), equal_to(3))

    @patch_rest
    def test_lun_snap_create_existing(self):
        lun = UnityLun(_id='sv_9', cli=t_rest())
        assert_that(lambda: lun.create_snap(name='lun_snap_1'),
                    raises(UnitySnapNameInUseError))

    @patch_rest
    def test_lun_rename(self):
        def f():
            lun = UnityLun(_id='sv_2', cli=t_rest())
            lun.name = 'Europa'

        assert_that(f, raises(UnityLunNameInUseError, 'already exists'))

    @patch_rest
    def test_lun_max_iops_property(self):
        lun = UnityLun(_id='sv_10', cli=t_rest())
        assert_that(lun.max_iops, equal_to(3600))
        assert_that(lun.max_kbps, none())

    @patch_rest
    def test_lun_max_kbps_property(self):
        lun = UnityLun(_id='sv_11', cli=t_rest())
        assert_that(lun.max_iops, none())
        assert_that(lun.max_kbps, equal_to(11000))

    @patch_rest
    def test_create_with_io_limit(self):
        cli = t_rest()
        policy = UnityIoLimitPolicy('qp_4', cli=cli)
        pool = UnityPool('pool_1', cli=cli)
        lun = pool.create_lun('Himalia', io_limit_policy=policy)
        assert_that(lun.name, equal_to('Himalia'))
        assert_that(lun.io_limit_policy.get_id(), equal_to('qp_4'))
        rule = lun.io_limit_rule
        assert_that(rule, instance_of(UnityIoLimitRuleSetting))
        assert_that(rule.max_kbps_density, equal_to(1100))
        assert_that(rule.name, equal_to('Density_1100_KBPS_rule'))

    @patch_rest
    def test_expand_lun_success(self):
        lun = UnityLun('sv_2', cli=t_rest())
        original_size = lun.expand(101 * 1024 ** 3)
        assert_that(original_size / 1024 ** 3, equal_to(100))

    @patch_rest
    def test_expand_lun_too_small(self):
        def f():
            lun = UnityLun('sv_2', cli=t_rest())
            lun.total_size_gb = 1

        assert_that(f, raises(UnityLunShrinkNotSupportedError, 'shrink'))

    @patch_rest
    def test_expand_lun_equal_size(self):
        def f():
            lun = UnityLun('sv_2', cli=t_rest())
            lun.total_size_gb = 100

        assert_that(f, raises(UnityNothingToModifyError, 'nothing to modify'))

    @patch_rest
    def test_lun_read_iops(self):
        lun = t_unity().get_lun(_id='sv_2')
        assert_that(lun.read_iops, equal_to(1.5))

    @patch_rest
    def test_lun_write_iops(self):
        lun = t_unity().get_lun(_id='sv_2')
        assert_that(lun.write_iops, equal_to(3.0))

    @patch_rest
    def test_lun_perf_disabled_exception(self):
        unity = UnitySystem('10.244.223.61', 'a', 'a')
        unity.disable_perf_stats()

        def f():
            return unity.get_lun(_id='sv_2').read_iops

        assert_that(f, raises(UnityPerfMonNotEnabledError, 'not enabled'))

    @patch_rest
    def test_thin_clone(self):
        lun = UnityLun.get(_id='sv_2', cli=t_rest(version='4.2.0'))
        clone = lun.thin_clone(name='test_thin_clone',
                               description='This is description.',
                               io_limit_policy=None)
        assert_that(clone.id, equal_to('sv_4678'))

    @patch_rest
    def test_thin_clone_limit_exceeded(self):
        lun = UnityLun.get(_id='sv_2', cli=t_rest(version='4.2.0'))

        def _inner():
            lun.thin_clone(name='test_thin_clone_limit_exceeded',
                           description='This is description.',
                           io_limit_policy=None)

        assert_that(_inner, raises(UnityThinCloneLimitExceededError))

    @patch_rest
    def test_update_hosts(self):
        lun = UnityLun.get(cli=t_rest(), _id="sv_4")
        r = lun.update_hosts(host_names=["ubuntu-server7"])
        assert_that(r, instance_of(RestResponse))

    @patch_rest
    def test_update_hosts_no_change(self):
        lun = UnityLun.get(cli=t_rest(), _id="sv_4")
        r = lun.update_hosts(host_names=["10.244.209.90"])
        assert_that(r, none())

    @patch_rest
    def test_migrate_lun_success(self):
        lun = UnityLun('sv_2', cli=t_rest())
        dest_pool = UnityPool('pool_4', cli=t_rest())
        r = lun.migrate(dest_pool)
        assert_that(r, equal_to(True))

    @patch_rest
    def test_migrate_compressed_lun_success(self):
        lun = UnityLun('sv_18', cli=t_rest())
        dest_pool = UnityPool('pool_5', cli=t_rest())
        r = lun.migrate(dest_pool)
        assert_that(r, equal_to(True))

    @patch_rest
    def test_migrate_lun_source_is_thin_clone(self):
        lun = UnityLun('sv_5606', cli=t_rest())
        dest_pool = UnityPool('pool_4', cli=t_rest())
        r = lun.migrate(dest_pool)
        assert_that(r, equal_to(False))

    @patch_rest
    def test_migrate_lun_source_compressed_dest_not_flash(self):
        lun = UnityLun('sv_18', cli=t_rest())
        dest_pool = UnityPool('pool_4', cli=t_rest())
        r = lun.migrate(dest_pool)
        assert_that(r, equal_to(False))

    @patch_rest
    def test_migrate_lun_has_thin_clone(self):
        lun = UnityLun('sv_2', cli=t_rest())
        dest_pool = UnityPool('pool_6', cli=t_rest())
        r = lun.migrate(dest_pool)
        assert_that(r, equal_to(False))

    @patch_rest
    def test_migrate_lun_pool_does_not_exist(self):
        lun = UnityLun('sv_2', cli=t_rest())
        dest_pool = UnityPool('pool_does_not_exist', cli=t_rest())
        assert_that(calling(lun.migrate).with_args(dest_pool),
                    raises(UnityMigrationSourceDestNotExistsError))

    @patch_rest
    def test_migrate_lun_failed(self):
        lun = UnityLun('sv_2', cli=t_rest())
        dest_pool = UnityPool('pool_7', cli=t_rest())
        r = lun.migrate(dest_pool)
        assert_that(r, equal_to(False))

    @patch_rest
    def test_migrate_lun_cancelled(self):
        lun = UnityLun('sv_2', cli=t_rest())
        dest_pool = UnityPool('pool_8', cli=t_rest())
        r = lun.migrate(dest_pool)
        assert_that(r, equal_to(False))

    @patch_rest
    def test_migrate_lun_timeout(self):
        lun = UnityLun('sv_2', cli=t_rest())
        dest_pool = UnityPool('pool_5', cli=t_rest())
        assert_that(calling(lun.migrate).with_args(dest_pool, timeout=10),
                    raises(UnityMigrationTimeoutException))

    @patch_rest
    @ddt.data(
        {'rep_name': None, 'rep_existing_snaps': None, 'remote_system': None},
        {'rep_name': 'remote-sv_2498-sv_5',
         'rep_existing_snaps': False, 'remote_system': 'RS_4'},
    )
    @ddt.unpack
    def test_replicate(self, rep_name, rep_existing_snaps, remote_system):
        if remote_system:
            remote_system = UnityRemoteSystem(_id=remote_system, cli=t_rest())
        lun = UnityLun.get(cli=t_rest(), _id='sv_2498')
        rep_session = lun.replicate(
            'sv_5', 60, replication_name=rep_name,
            replicate_existing_snaps=rep_existing_snaps,
            remote_system=remote_system
        )
        assert_that(rep_session.name, equal_to('remote-sv_2498-sv_5'))

    @patch_rest
    @ddt.data(
        {'dst_lun_name': None, 'remote_system': None, 'rep_name': None,
         'dst_size': None, 'dst_sp': None, 'is_dst_thin': None,
         'dst_tiering_policy': None, 'is_dst_compression': None},
        {'dst_lun_name': 'lun-rep-src3-liangr', 'remote_system': 'RS_4',
         'rep_name': 'remote-rep3', 'dst_size': 10737418240,
         'dst_sp': NodeEnum.SPA, 'is_dst_thin': True,
         'dst_tiering_policy': TieringPolicyEnum.AUTOTIER_HIGH,
         'is_dst_compression': False},
    )
    @ddt.unpack
    def test_replicate_with_dst_resource_provisioning(self,
                                                      dst_lun_name,
                                                      remote_system,
                                                      rep_name,
                                                      dst_size,
                                                      dst_sp,
                                                      is_dst_thin,
                                                      dst_tiering_policy,
                                                      is_dst_compression):
        lun = UnityLun.get(cli=t_rest(), _id='sv_1876')
        if remote_system:
            remote_system = UnityRemoteSystem(_id=remote_system, cli=t_rest())
        rep_session = lun.replicate_with_dst_resource_provisioning(
            60, 'pool_2', dst_lun_name=dst_lun_name,
            remote_system=remote_system, replication_name=rep_name,
            dst_size=dst_size, dst_sp=dst_sp, is_dst_thin=is_dst_thin,
            dst_tiering_policy=dst_tiering_policy,
            is_dst_compression=is_dst_compression)
        assert_that(rep_session.id, equal_to(
            '42949675780_FNM00150600267_0000_42949678642_FNM00152000052_0000'))

    @patch_rest
    def test_is_vmware_vmfs_false(self):
        vmfs = UnityLun.get(cli=t_rest(), _id='sv_4')
        assert_that(vmfs.is_vmware_vmfs, equal_to(False))

    @patch_rest
    def test_is_vmware_vmfs_true(self):
        vmfs = UnityLun.get(cli=t_rest(), _id='sv_5613')
        assert_that(vmfs.is_vmware_vmfs, equal_to(True))

    @patch_rest
    def test_modify_vmfs_name_major_version_block_size(self):
        vmfs = UnityLun.get(cli=t_rest(), _id='sv_5613')
        resp = vmfs.modify(name='vmfs_new_name', sp=1,
                           major_version=ESXFilesystemMajorVersionEnum.VMFS_6,
                           block_size=ESXFilesystemBlockSizeEnum._4MB)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_delete_vmfs(self):
        vmfs = UnityLun.get(cli=t_rest(), _id='sv_5613')
        resp = vmfs.delete(async_mode=False)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_create_with_snap_schedule(self):
        cli = t_rest()
        schedule = UnitySnapSchedule(_id='snapSch_1', cli=cli)
        pool = UnityPool('pool_1', cli=cli)
        lun = pool.create_lun('lun-with-snap-schedule', snap_schedule=schedule)
        assert_that(lun.name, equal_to('lun-with-snap-schedule'))
        assert_that(lun.snap_schedule.get_id(), equal_to('snapSch_1'))

    @patch_rest
    def test_modify_snap_schedule(self):
        cli = t_rest()
        new_schedule = UnitySnapSchedule(_id='snapSch_4', cli=cli)
        lun = UnityLun(_id='sv_16455', cli=cli)
        resp = lun.modify(snap_schedule=new_schedule)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_remove_snap_schedule(self):
        cli = t_rest()
        lun = UnityLun(_id='sv_16455', cli=cli)
        resp = lun.remove_snap_schedule()
        assert_that(resp.is_ok(), equal_to(True))


class UnityLunEnablePerfStatsTest(TestCase):
    @patch_rest
    def setUp(self):
        self.unity = UnitySystem('10.244.223.61', 'a', 'a')
        self.unity.enable_perf_stats(1, [UnityDisk])

    @patch_rest
    def tearDown(self):
        self.unity.disable_perf_stats()

    @patch_rest
    def test_lun_perf_not_enabled_exception(self):
        disk = self.unity.get_disk(_id='dae_0_1_disk_0')
        assert_that(disk.read_iops, is_nan())

        def f():
            return self.unity.get_lun(_id='sv_2').read_iops

        assert_that(f, raises(UnityPerfMonNotEnabledError, 'not enabled'))

    @patch_rest
    def test_lun_properties_perf_not_enabled(self):
        lun = self.unity.get_lun(_id='sv_2')
        assert_that(lun.property_names, is_not(has_item('read_iops')))
        disk = self.unity.get_disk(_id='dae_0_1_disk_0')
        assert_that(disk.property_names(), has_item('read_iops'))

    @patch_rest
    def test_nested_properties(self):
        lun = self.unity.get_lun(_id='sv_12')
        assert_that(lun.pool.raid_type, equal_to(RaidTypeEnum.RAID10))
        assert_that(lun.pool.is_fast_cache_enabled, equal_to(False))
        assert_that(lun.host_access[0].host.name,
                    equal_to('Virtual_Machine_12'))

    @patch_rest
    def test_create_snap_of_member_snap_not_support(self):
        lun = UnityLun(cli=t_rest(), _id='sv_58')
        assert_that(calling(lun.create_snap).with_args(name='not-support'),
                    raises(UnityCGMemberActionNotSupportError))

    @patch_rest
    def test_thinclone_of_member_snap_not_support(self):
        lun = UnityLun(cli=t_rest(version='4.3'), _id='sv_58')
        assert_that(calling(lun.thin_clone).with_args('not-support'),
                    raises(UnityCGMemberActionNotSupportError))

    @patch_rest
    def test_thinclone_of_thick_lun_not_allowed(self):
        lun = UnityLun(cli=t_rest(version='4.3'), _id='sv_59')
        assert_that(calling(lun.thin_clone).with_args('not-allowed'),
                    raises(UnityThinCloneNotAllowedError))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import copy
import json
import logging
import time

import requests
import six
from requests.exceptions import RequestException
from retryz import retry

from storops.connection import exceptions
from storops.exception import StoropsConnectTimeoutError

log = logging.getLogger(__name__)


def _on_error_callback(e):
    return isinstance(e, RequestException)


def _wait_callback(tried):
    return 2 ** (tried - 1)


class HTTPClient(object):
    def __init__(self, base_url, headers, insecure=False, auth=None,
                 timeout=None, retries=None, ca_cert_path=None,
                 cache_interval=0):
        self.base_url = base_url
        if retries is None:
            retries = 2
        self.retries = retries
        self.request_options = self._set_request_options(
            insecure, auth, timeout, ca_cert_path)
        self.headers = headers
        self.session = requests.session()
        self.cache_interval = cache_interval
        self.cache_dict = {}

    def __del__(self):
        self.session.close()

    @staticmethod
    def _set_request_options(insecure=None, auth=None, timeout=None,
                             ca_cert_path=None):
        options = {'verify': True}

        if ca_cert_path is not None:
            options['verify'] = ca_cert_path

        if insecure:
            options['verify'] = False

        if timeout:
            options['timeout'] = timeout

        options['auth'] = auth

        return options

    def request(self, full_url, method, **kwargs):
        headers = copy.deepcopy(self.headers)
        headers.update(kwargs.get('headers', {}))
        options = copy.deepcopy(self.request_options)
        content_type = headers.get('Content-Type', None)
        if kwargs.get('body', None):
            if content_type == 'application/json':
                options['data'] = json.dumps(kwargs['body'])
            else:
                options['data'] = kwargs['body']
        files = kwargs.get('files', None)
        files_opener = None
        if files:
            files_opener = {}
            headers.pop('Content-Type')
            for name, path in files.items():
                files_opener[name] = open(path, 'rb')
        self.log_request(full_url, method, options.get('data', None))
        start = time.time()
        resp = self.session.request(method, full_url, headers=headers,
                                    files=files_opener, **options)

        self.log_response(full_url, method, resp, start)

        body = None
        if resp.text:
            try:
                if content_type == 'application/json':
                    body = json.loads(resp.text)
                else:
                    body = resp.text

            except ValueError:
                pass

        # These errors should be raised directly, do NOT retry any more,
        # like 401 - unauthorized, 503 - service unavailable.
        if resp.status_code in [401, 503]:
            raise exceptions.from_response(resp, method, full_url)

        return resp, body

    def _cs_request(self, url, method, **kwargs):
        try:
            return self._cs_request_with_retries(
                self.base_url + url,
                method,
                **kwargs)
        except requests.ConnectTimeout as ex:
            raise StoropsConnectTimeoutError(message=str(ex))

    def _get_limit(self):
        return self.retries

    @retry(wait=_wait_callback, limit=_get_limit,
           on_error=_on_error_callback)
    def _cs_request_with_retries(self, url, method, **kwargs):
        return self.request(url, method, **kwargs)

    def get(self, url, **kwargs):
        if self.cache_interval <= 0:
            result = self._cs_request(url, 'GET', **kwargs)
        else:
            current = time.time()
            if url in self.cache_dict.keys() and (
                    current - self.cache_dict.get(url)[
                    0] < self.cache_interval):
                result = self.cache_dict[url][1]
                log.debug('Read response from cache, URL: {}'.format(url))
            else:
                result = self._cs_request(url, 'GET', **kwargs)
                current = time.time()
                log.debug('Write response to cache, URL: {}'.format(url))
                self.cache_dict[url] = (current, result)
        return result

    def post(self, url, **kwargs):
        return self._cs_request(url, 'POST', **kwargs)

    def put(self, url, **kwargs):
        return self._cs_request(url, 'PUT', **kwargs)

    def delete(self, url, **kwargs):
        return self._cs_request(url, 'DELETE', **kwargs)

    @classmethod
    def log_request(cls, url, method, data=None):
        if log.isEnabledFor(logging.DEBUG):
            log.debug('REQ URL: [{}] {}'.format(method, url))
            cls._debug_print_json(data, 'REQ BODY:')

    @staticmethod
    def _debug_print_json(data, prefix=None):
        if prefix is None:
            prefix = 'JSON data:'

        if data is not None and log.isEnabledFor(logging.DEBUG):
            try:
                if isinstance(data, six.string_types):
                    obj = json.loads(data, encoding='utf-8')
                else:
                    obj = data
                text = json.dumps(obj, indent=4)
            except (ValueError, TypeError):
                text = data
            log.debug('{}\n{}'.format(prefix, text))

    @classmethod
    def log_response(cls, full_url, method, resp, start_time):
        if log.isEnabledFor(logging.DEBUG):
            dt = time.time() - start_time
            log.debug('REQ URL: [{}] {}, TIME: {}, RESP CODE: {}'
                      .format(method, full_url, dt, resp.status_code))
            cls._debug_print_json(resp.text, 'RESP BODY:')

    def update_headers(self, headers):
        self.headers.update(headers)



# coding=utf-8
# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import threading
import time

import persistqueue
from persistqueue import Empty

from storops.exception import StoropsException
from storops.exception import VNXObjectNotFoundError, \
    UnityResourceNotFoundError

__author__ = 'Peter Wang'

log = logging.getLogger(__name__)


class PQueue(object):
    DEFAULT_INTERVAL = 300
    MAX_RETRIES = 100

    def __init__(self, path, interval=None):
        self.path = path
        self._q = persistqueue.Queue(self.path, tempdir=self.path)
        self._interval = (
            self.DEFAULT_INTERVAL if interval is None else interval)
        self.started = False

    def put(self, func, **kwargs):
        item = {'object': func.__self__, 'method': func.__name__,
                'params': kwargs}
        self._q.put_nowait(item)

    def get(self, block=False):
        return self._q.get(block=block)

    def task_done(self):
        self._q.task_done()

    def start(self):
        if not self.started:
            self._run()
            self.started = True
        else:
            log.info("PQueue[{}] had already started.".format(self.path))

    def stop(self):
        self._interval = 0
        self.started = False

    def set_interval(self, interval):
        self._interval = interval

    def _run(self):
        self._t = threading.Thread(target=self._run_tasks)
        self._t.setDaemon(True)
        self._t.start()

    def re_enqueue(self, item):
        """Re-enqueue till reach max retries."""
        if 'retries' in item:
            retries = item['retries']
            if retries >= self.MAX_RETRIES:
                log.warn("Failed to execute {} after {} retries, give it "
                         " up.".format(item['method'], retries))
            else:
                retries += 1
                item['retries'] = retries
                self._q.put_nowait(item)
        else:
            item['retries'] = 1
            self._q.put_nowait(item)

    @staticmethod
    def _is_ignored_exception(ex):
        ignored_exceptions = (VNXObjectNotFoundError,
                              UnityResourceNotFoundError)
        for ignored in ignored_exceptions:
            if isinstance(ex, ignored):
                return True
        return False

    @staticmethod
    def _is_reenqueue_exception(ex):
        reenqueue_exceptions = (StoropsException, )
        for reenqueue in reenqueue_exceptions:
            if isinstance(ex, reenqueue):
                return True
        return False

    def _run_tasks(self):
        while self._interval > 0:
            log.debug("Running periodical check.")
            try:
                data = self.get()
            except Empty:
                log.debug("Queue is empty now.")
            else:
                method = getattr(data['object'], data['method'], None)
                try:
                    method(**data['params'])
                except Exception as ex:
                    log.debug("Failed to execute {}: {}, this message can be "
                              "safely ignored.".format(method.__name__,
                                                       ex))
                    if self._is_ignored_exception(ex):
                        log.info("Object had been deleted: {}, this message "
                                 "can be safely ignored.".format(ex))
                    elif self._is_reenqueue_exception(ex):
                        # Re-enqueue since failed to execute
                        self.re_enqueue(data)
                    else:
                        log.error("Unexpected error occurs when executing {}:"
                                  " {}, this job will not be executed"
                                  " again".format(method.__name__, ex))
                self.task_done()
            time.sleep(self._interval)
        log.info("{} with path {} has been "
                 "stopped.".format(self.__class__.__name__, self._q.path))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging

import storops.vnx.resource.fs
import storops.vnx.resource.nas_pool
from storops.vnx.resource import VNXCliResourceList, VNXResource

__author__ = 'Jay Xu'

log = logging.getLogger(__name__)


class VNXFsSnapList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXFsSnap

    def _get_raw_resource(self):
        return self._cli.get_fs_snap()


class VNXFsSnap(VNXResource):
    def __init__(self, name=None, snap_id=None, cli=None):
        super(VNXFsSnap, self).__init__()
        self._name = name
        self._snap_id = snap_id
        self._cli = cli

    def _get_raw_resource(self):
        if self._name is not None or self._snap_id is not None:
            ret = self._cli.get_fs_snap(self._name, snap_id=self._snap_id)
        else:
            raise ValueError('snap name should be specified.')
        return ret

    @staticmethod
    def get(cli=None, name=None, snap_id=None):
        if name is not None or snap_id is not None:
            ret = VNXFsSnap(name=name, snap_id=snap_id, cli=cli)
        else:
            ret = VNXFsSnapList(cli=cli)
        return ret

    def get_snap_id(self):
        if self._snap_id is not None:
            ret = self._snap_id
        else:
            ret = self.snap_id
        return ret

    @property
    def fs(self):
        clz = storops.vnx.resource.fs.VNXFileSystem
        return clz(fs_id=self.fs_id, cli=self._cli)

    @classmethod
    def create(cls, cli, name, fs, pool, size=None):
        fs_clz = storops.vnx.resource.fs.VNXFileSystem
        pool_clz = storops.vnx.resource.nas_pool.VNXNasPool
        fs_id = fs_clz.get_id(fs)
        pool_id = pool_clz.get_id(pool)
        resp = cli.create_snap(name, fs_id, pool_id, size)
        resp.raise_if_err()
        return VNXFsSnap(name=name, cli=cli)

    def delete(self):
        return self._cli.delete_snap(self.get_snap_id())



# coding=utf-8
# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from storops.exception import raise_if_err, VNXStatsException
from storops.vnx.resource import VNXCliResource

__author__ = 'Cedric Zhuang'


class VNXStats(VNXCliResource):
    def __init__(self, cli=None):
        super(VNXStats, self).__init__()
        self._cli = cli

    @classmethod
    def get(cls, cli):
        return VNXStats(cli=cli)

    def is_enabled(self):
        out = self._cli.set_stats()
        if 'ENABLED' in out:
            ret = True
        elif 'DISABLED' in out:
            ret = False
        else:
            raise VNXStatsException(out)
        return ret

    def enable_stats(self):
        out = self._cli.set_stats(True)
        raise_if_err(out, VNXStatsException)

    def disable_stats(self):
        out = self._cli.set_stats(False)
        raise_if_err(out, VNXStatsException)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import unittest

from hamcrest import assert_that, is_not, equal_to
from hamcrest import contains_string

from storops_test.unity.rest_mock import MockRestClient

__author__ = 'Cedric Zhuang'


class TestMockRestClient(unittest.TestCase):
    def test_get_folder_from_url(self):
        url = '/api/types/metric?compact=true&fields=name,type'
        folder = MockRestClient.get_folder_from_url(url)
        assert_that(folder, is_not(contains_string('api')))
        assert_that(folder, is_not(contains_string('types')))
        assert_that(folder, contains_string('metric'))

    def test_get_folder_from_instance_url(self):
        url = '/api/instances/lun/sv_2'
        folder = MockRestClient.get_folder_from_url(url)
        assert_that(folder, is_not(contains_string('api')))
        assert_that(folder, is_not(contains_string('types')))
        assert_that(folder, contains_string('lun'))

    def test_get_filename(self):
        inputs = ['/api/instances/lun/sv_2?compact=True&fields=type,wwn', None]
        name = MockRestClient.get_filename(inputs)
        assert_that(name, equal_to('for_test.json'))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import glob
import inspect
import logging
import os
import re

import six

import yaml
from yaml import loader

from storops.lib.common import cache, instance_cache, Enum, \
    get_clz_from_module, EnumList
from storops.lib import converter as cvt
import storops.lib.resource

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


class PropMapper(object):
    def __init__(self,
                 label,
                 key=None,
                 option=None,
                 converter=None):
        self.option = option
        self._label = label
        self._key = key
        self.converter = converter

    @property
    def key(self):
        if self._key is None:
            if self.label is None:
                raise ValueError('"label" not found!')
            self._key = self.camel_case_to_under_score(self.label).lower()
        return self._key

    to_delete = re.compile(r'[\(\)]')
    p0 = re.compile(r"[A-Za-z0-9']+")
    p1 = re.compile(r"([^_])([A-Z]s[a-z]+|[A-Z][a-z]{2})")
    p2 = re.compile(r'([a-z0-9])([A-Z])')

    @classmethod
    def camel_case_to_under_score(cls, value, delimiter='_'):
        value = re.sub(cls.to_delete, '', value)
        value = '_'.join(re.findall(cls.p0, value))
        s1 = re.sub(cls.p1, r'\1_\2', value)
        ret = re.sub(cls.p2, r'\1_\2', s1).lower()
        if delimiter != '_':
            ret = ret.replace('_', delimiter)
        return ret

    @property
    def label(self):
        return self._label.strip()


class PropDescriptor(PropMapper):
    def __init__(self,
                 option,
                 label,
                 key=None,
                 converter=None,
                 is_index=False,
                 end_pattern=None,
                 is_regex=False):
        super(PropDescriptor, self).__init__(label,
                                             key=key,
                                             option=option,
                                             converter=converter)
        self._is_index = is_index
        self._end_pattern = None
        self._is_regex = is_regex
        self.sequence = -1

        # properties setter
        self.end_pattern = end_pattern

        # other initializations
        self._validate_config()

    def _validate_config(self):
        if self.is_regex and self.end_pattern is not None:
            raise ValueError('"is_regex" and "end_pattern" '
                             'are mutual exclusive.')

    @property
    def is_regex(self):
        return self._is_regex

    @property
    def is_index(self):
        return self._is_index

    def convert(self, value):
        c = self.converter
        ret = value
        if c is not None:
            if self.is_parser():
                ret = c.parse_all(value)
            elif self.is_resource_clazz():
                # value is raw output from cli.
                ret = c().update(value)
            elif self.is_enum() or self.is_enum_list():
                ret = c.parse(value)
            elif callable(c):
                ret = c(value)
        return ret

    def is_resource_list_clazz(self):
        rsc_list_clz = storops.lib.resource.ResourceList
        c = self.converter
        return inspect.isclass(c) and issubclass(c, rsc_list_clz)

    def is_resource_clazz(self):
        rsc_clz = storops.lib.resource.Resource
        c = self.converter
        return inspect.isclass(c) and issubclass(c, rsc_clz)

    def is_enum(self):
        c = self.converter
        return inspect.isclass(c) and issubclass(c, Enum)

    def is_enum_list(self):
        c = self.converter
        return inspect.isclass(c) and issubclass(c, EnumList)

    def is_parser(self):
        c = self.converter
        return isinstance(c, type) and issubclass(c, OutputParser)

    @property
    def end_pattern(self):
        return self._end_pattern

    @end_pattern.setter
    def end_pattern(self, value):
        if isinstance(value, six.string_types) or value is None:
            self._end_pattern = value
        else:
            raise ValueError('value must be a string type.')

    @property
    @instance_cache
    def index_pattern(self):
        if not self.is_regex:
            ret = re.compile(
                r"(^\s*{})".format(re.escape(self.label)),
                flags=self.re_flags)
        else:
            ret = re.compile(self.label, flags=self.re_flags)
        return ret

    re_flags = re.MULTILINE | re.IGNORECASE

    @property
    @instance_cache
    def pattern(self):
        if self.end_pattern is None:
            if self.is_regex:
                ret = re.compile(self.label, self.re_flags)
            else:
                ret = re.compile(
                    '^[ \t]*{}[ \t]*(?P<value>.*)[ \t]*$'.format(
                        re.escape(self.label)), self.re_flags)
        else:
            # has 'end_pattern' means we need to match
            # multi-lines including return.
            # do NOT escape 'end_pattern' because it's
            # already a PATTERN.
            self.re_flags |= re.DOTALL
            ret = re.compile(
                r'^\s*{}\s*(?P<value>.*){}'.format(
                    re.escape(self.label),
                    self.end_pattern),
                self.re_flags)
        return ret

    def __lt__(self, other):
        if isinstance(other, PropDescriptor):
            ret = self.sequence < other.sequence
        else:
            ret = True
        return ret


class OutputParser(object):
    def __init__(self):
        self._property_map = {}
        self.resource_name = ''

    @property
    def data_src(self):
        raise NotImplementedError('must be implemented by sub-class')

    @property
    def properties(self):
        if self._property_map is None:
            self._property_map = {}
        return self._property_map.values()

    def set_property_map(self, value):
        self._property_map = value

    def add_property(self, *props):
        for prop in props:
            seq = len(self._property_map)
            prop.sequence = seq
            self._property_map[prop.key.upper()] = prop

    def has_property_key(self, key):
        return key.upper() in self._property_map

    def get_property(self, key):
        return self._property_map.get(key.upper(), None)

    def get_property_label(self, key):
        prop = self.get_property(key)
        if prop:
            ret = prop.label
        else:
            ret = None
        return ret

    def get_property_key(self, label):
        for k, v in self._property_map.items():
            if v.label == label:
                ret = k.lower()
                break
        else:
            ret = None
        return ret

    @property
    def property_names(self):
        return sorted(p.key for p in self.properties)

    def parse_all(self, output, properties=None):
        raise NotImplementedError('must be implemented by sub-class')

    def parse(self, output, properties=None):
        raise NotImplementedError('must be implemented by sub-class')

    @property
    def all_options(self):
        return sorted(d.option for d in self.properties
                      if d.option is not None)

    def init_from_config(self, config):
        # do any customized initialization from config in child class
        pass

    @property
    @instance_cache
    def index_property(self):
        indices = self.index_property_list
        # make sure only one index for each parser
        if len(indices) == 1:
            ret = indices[0]
        elif len(indices) > 1:
            # find the first index
            indices.sort(key=lambda i: i.sequence)
            ret = indices[0]
            log.info('multiple index configured.  '
                     'use the first one as the splitter of output.  '
                     'first index is: "{}"'.format(ret.label))
        else:
            ret = None

        return ret

    @property
    def index_property_list(self):
        return sorted(p for p in self.properties
                      if p.is_index is True)

    @property
    def property_options(self):
        return sorted(p.option for p in self.properties
                      if p.option is not None)

    def __getattr__(self, item):
        if item in self._property_map.keys():
            ret = self._property_map[item]
        else:
            ret = super(OutputParser, self).__getattribute__(item)
        return ret


class ParserConfigFactory(object):
    config_filename = 'parser_configs.yaml'

    @classmethod
    def get_parser_clz(cls, data_src):
        raise NotImplementedError('get_base_clz not implemented.')

    @classmethod
    def get_converter(cls, _):
        return None

    @classmethod
    def get_folder(cls):
        return os.path.dirname(inspect.getfile(cls))

    @classmethod
    def get_rsc_pkg_name(cls):
        names = cls.__module__.split('.')[:-1]
        names.append('resource')
        return '.'.join(names)

    def get(self, name):
        config = self.get_config(name)

        parser = self._get_parser_instance(config)
        parser.init_from_config(config)

        self._get_property_map(parser, config)
        parser.resource_class_name = name
        return parser

    def _get_parser_instance(self, config):
        base_clz = self.get_parser_clz(config.data_src)
        parser = base_clz()
        return parser

    def _get_property_map(self, parser, config):
        for v in config.properties:
            p = self.init_descriptor(v)
            parser.add_property(p)

    def get_config(self, name):
        all_configs = self._read_configs()
        if name not in all_configs:
            raise ValueError('cannot find {} in {}.'.format(
                name, self.config_filename))
        return ParserConfig(all_configs[name])

    def _read_configs(self):
        filename = os.path.join(self.get_folder(), self.config_filename)
        with open(filename, 'r') as stream:
            ret = yaml.load(stream, Loader=loader.SafeLoader)
        return ret

    @instance_cache
    def get_resource_clz_by_name(self, clz_name):
        ret = None
        if isinstance(clz_name, six.string_types):
            sub_module_names = self._rsc_sub_module_names()
            for sub_module_name in sub_module_names:
                ret = get_clz_from_module(sub_module_name, clz_name)
                if ret is not None:
                    break
        return ret

    @instance_cache
    def get_enum_by_name(self, clz_name):
        ret = None
        if isinstance(clz_name, six.string_types):
            enum_module_name = self._get_enum_module_name()
            ret = get_clz_from_module(enum_module_name, clz_name)
        return ret

    def _get_enum_module_name(self):
        names = self.__module__.split('.')[:-1]
        names.append('enums')
        return '.'.join(names)

    @cache
    def _rsc_sub_module_names(self):
        resource_folder = os.path.join(self.get_folder(), 'resource')
        resource_files = glob.glob('{}{}*.py'.format(resource_folder, os.sep))
        ret = []
        for path in resource_files:
            pkg_name = os.path.basename(path).split('.')[0]
            full_pkg_name = '.'.join([self.get_rsc_pkg_name(), pkg_name])
            ret.append(full_pkg_name)
        return ret

    def _get_converter(self, converter_str):
        """find converter function reference by name

        find converter by name, converter name follows this convention:

            Class.method

        or:

            method

        The first type of converter class/function must be available in
        current module.
        The second type of converter must be available in `__builtin__`
        (or `builtins` in python3) module.

        :param converter_str: string representation of the converter func
        :return: function reference
        """
        ret = None
        if converter_str is not None:
            converter_desc_list = converter_str.split('.')
            if len(converter_desc_list) == 1:
                converter = converter_desc_list[0]
                # default to `converter`
                ret = getattr(cvt, converter, None)

                if ret is None:
                    # try module converter
                    ret = self.get_converter(converter)

                if ret is None:
                    ret = self.get_resource_clz_by_name(converter)

                if ret is None:
                    ret = self.get_enum_by_name(converter)

                if ret is None:
                    # try parser config
                    ret = self.get(converter)

            if ret is None and converter_str is not None:
                raise ValueError(
                    'Specified converter not supported: {}'.format(
                        converter_str))
        return ret

    def init_descriptor(self, prop):
        converter = self._get_converter(prop.get('converter', None))
        return PropDescriptor(option=prop.get('option', None),
                              label=prop.get('label', None),
                              key=prop.get('key', None),
                              is_index=prop.get('is_index', None),
                              converter=converter,
                              end_pattern=prop.get('end_pattern', None),
                              is_regex=prop.get('is_regex', None))


class ParserConfig(object):
    def __init__(self, inputs):
        self.data_src = inputs.get('data_src', None)
        self.name = inputs.get('name', None)
        self._properties = inputs.get('properties', None)
        if self._properties is None:
            self._properties = []

    @property
    def properties(self):
        return self._properties

    def add_property(self, prop):
        self._properties.append(prop)



# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, equal_to

from storops.unity.enums import SeverityEnum
from storops.unity.resource.alert import UnityAlert
from storops_test.unity.rest_mock import t_rest, patch_rest


class UnityAlertConfigSNMPTargetTest(TestCase):
    @patch_rest
    def test_alert_properties(self):
        alert = UnityAlert(_id='alert_180', cli=t_rest())
        assert_that(alert.id, equal_to('alert_180'))
        assert_that(alert.severity, equal_to(SeverityEnum.CRITICAL))
        assert_that(alert.instance_id, equal_to(
            'root/emc:C4CB_AlertEvent%AlertID=1577836740661/637773388'))
        assert_that(alert.timestamp, equal_to('2019-12-31T23:59:00.661Z'))
        assert_that(alert.message_id, equal_to('14:170002'))
        assert_that(alert.arguments, equal_to(['BASE_OE_V4_0']))
        assert_that(alert.catalog_id, equal_to('CEM-CatalogService'))
        assert_that(alert.message, equal_to(
            "The BASE_OE_V4_0 license has expired, and the storage system's "
            "support for the licensed feature has been disabled. Obtain and "
            "install a new license file to ensure support for the licensed "
            "feature."))
        assert_that(alert.description_id, equal_to(
            'ALRT_LICENSINGSERVICES_LICENSE_EXPIRED'))
        assert_that(alert.description, equal_to(
            "One of your system licenses has expired or will expire soon. "
            "Obtain and install the license file to ensure continued access "
            "to the relevant feature."))
        assert_that(alert.resolution_id, equal_to('licnse_update'))
        assert_that(alert.resolution, equal_to(
            '/help/webhelp/en_US/index.html?#unity_t_manage_licenses.html'))
        assert_that(alert.component_facility, equal_to('LicensingServices'))
        assert_that(alert.is_acknowledged, equal_to(False))

    @patch_rest
    def test_modify_alert(self):
        alert = UnityAlert(_id='alert_181', cli=t_rest())
        resp = alert.modify(is_acknowledged=True)
        alert.update()
        assert_that(resp.is_ok(), equal_to(True))
        assert_that(alert.is_acknowledged, equal_to(True))

    @patch_rest
    def test_delete_alert(self):
        alert = UnityAlert(_id='alert_182', cli=t_rest())
        resp = alert.delete()
        assert_that(resp.body, equal_to({}))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from datetime import datetime
from unittest import TestCase

from hamcrest import assert_that, equal_to, none, has_items, instance_of, \
    only_contains, raises

from storops.lib import converter
from storops.vnx.enums import VNXRaidType

__author__ = 'Cedric Zhuang'


class ConverterTest(TestCase):
    def test_to_int_arr(self):
        ret = converter.to_int_arr('12, 0, 5, 12')
        assert_that(ret, equal_to([12, 0, 5, 12]))

    def test_to_int_arr_space(self):
        ret = converter.to_int_arr('62 63 306 324')
        assert_that(ret, equal_to([62, 63, 306, 324]))

    def test_to_int_arr_space_str(self):
        ret = converter.to_int_arr('Unbound')
        assert_that(ret, equal_to([]))

    def test_to_wwn(self):
        ret = converter.to_wwn('ab1234')
        assert_that(ret, equal_to('AB:12:34'))

    def test_to_wwn_no_change(self):
        ret = converter.to_wwn('ab:12:34')
        assert_that(ret, equal_to('AB:12:34'))

    def test_to_wwn_not_aligned(self):
        ret = converter.to_wwn('ab123')
        assert_that(ret, equal_to('AB:12:3'))

    def test_to_int_arr_empty_input(self):
        ret = converter.to_int_arr('')
        assert_that(ret, equal_to([]))

    def test_to_int_str_map(self):
        ret = converter.to_int_str_map(
            '62: RAID5 63: RAID5 306: RAID10 324: RAID5')
        assert_that(len(ret), equal_to(4))
        assert_that(ret[62], equal_to('RAID5'))
        assert_that(ret[306], equal_to('RAID10'))

    def test_to_int_str_map_na(self):
        ret = converter.to_int_str_map('N/A')
        assert_that(len(ret), equal_to(0))

    def test_to_int_int_map_unbound(self):
        ret = converter.to_int_int_map('Unbound')
        assert_that(len(ret), equal_to(0))

    def test_to_int_int_map(self):
        ret = converter.to_int_int_map('62: 92 63: 93 306: 94 324: 95')
        assert_that(len(ret), equal_to(4))
        assert_that(ret[62], equal_to(92))
        assert_that(ret[324], equal_to(95))

    def test_str_to_int_invalid_input(self):
        ret = converter.to_int_arr('12, abc, 12c')
        assert_that(ret, only_contains(12))

    def test_arr_to_str(self):
        ret = converter.arr_to_str([5, -12, 7.21])
        assert_that(ret, equal_to('5,-12,7.21'))

    def test_arr_to_str_with_sep(self):
        ret = converter.arr_to_str(['a bc', 112], '|')
        assert_that(ret, equal_to('a bc|112'))

    def test_to_str_arr(self):
        ret = converter.to_str_arr([5, -12, 7.21])
        assert_that(ret, only_contains('5', '-12', '7.21'))

    def test_to_hlu_alu_map(self):
        output = """
                 HLU/ALU Pairs:

                 HLU Number     ALU Number
                 ----------     ----------
                    0               4
                    12              A1
                 """
        alu2hlu = converter.to_alu_hlu_map(output)
        assert_that(alu2hlu[4], equal_to(0))
        assert_that(alu2hlu.get('A1', None), none())

    def test_to_disk_indices(self):
        output = """
        Disks:
        Bus 0 Enclosure 0 Disk 9
        Bus 1 Enclosure 0 Disk 12
        Bus 1 Enclosure 0 Disk 9
        """
        indices = converter.to_disk_indices(output)
        assert_that(indices, has_items('0_0_9', '1_0_12', '1_0_9'))

    def test_number_to_date_int(self):
        ret = converter.vnx_time_to_date(3623881621)
        assert_that(str(ret), equal_to('2015-11-02 01:47:01'))

    def test_number_to_date_float(self):
        ret = converter.vnx_time_to_date(3617925585.15)
        assert_that(str(ret), equal_to('2015-08-25 03:19:45.150000'))

    def test_float_normal(self):
        ret = converter.to_float('1234.5678')
        assert_that(ret, equal_to(1234.5678))

    def test_float_na(self):
        ret = converter.to_float('N/A')
        assert_that(ret, none())

    def test_int_normal(self):
        ret = converter.to_float('1234')
        assert_that(ret, equal_to(1234))

    def test_to_int_arr_from_str_arr(self):
        ret = converter.to_int_arr(['12', '23'])
        assert_that(ret, has_items(12, 23))

    def test_int_disabled(self):
        ret = converter.to_int('Disabled')
        assert_that(ret, none())

    def test_to_float_normal(self):
        ret = converter.to_float('12.34')
        assert_that(ret, equal_to(12.34))

    def test_to_float_percent(self):
        ret = converter.to_float('12.34%')
        assert_that(ret, equal_to(12.34))

    def test_to_float_invalid(self):
        ret = converter.to_float('12.34.56')
        assert_that(ret, none())

    def test_to_hex(self):
        ret = converter.to_hex(13691781134)
        assert_that(ret, equal_to('0x33018000e'))

    def test_to_datetime(self):
        ret = converter.to_datetime('2016-03-02T02:43:34.014Z')
        assert_that(ret, instance_of(datetime))
        assert_that(str(ret), equal_to('2016-03-02 02:43:34.014000+00:00'))

    def test_to_time_delta_more_than_1_day(self):
        ret = converter.to_time_delta("31:02:34.567")
        assert_that(str(ret), equal_to('1 day, 7:02:34.567000'))

    def test_to_time_delta_zero(self):
        ret = converter.to_time_delta("00:00:00.000")
        assert_that(str(ret), equal_to('0:00:00'))

    def test_url_to_host(self):
        url = 'http://www.abc.com:5050/page.xml'
        ret = converter.url_to_host(url)
        assert_that(ret, equal_to('www.abc.com'))

    def test_url_to_host_no_host(self):
        url = 'http://'

        def expect_error():
            converter.url_to_host(url)
        assert_that(expect_error, raises(ValueError))

    def test_url_to_host_ssl(self):
        url = 'https://10.0.0.1:4443/page/my.html?filter=name'
        ret = converter.url_to_host(url)
        assert_that(ret, equal_to('10.0.0.1'))

    def test_parse_host_address_ipv6(self):
        ipv6 = '2001:db8:a0b:12f0::1%eth0'
        address, prefix, netmask = converter.parse_host_address(ipv6)
        assert_that(address, equal_to('2001:db8:a0b:12f0::1%eth0'))
        assert_that(prefix, equal_to(None))
        assert_that(netmask, equal_to(None))

    def test_url_to_host_default_port(self):
        url = 'https://[2001:db8:a0b:12f0::1%eth0]/my.txt'
        ret = converter.url_to_host(url)
        assert_that(ret, equal_to('[2001:db8:a0b:12f0::1%eth0]'))

    def test_url_to_host_default_protocol(self):
        url = '10.0.0.1/my.txt'
        ret = converter.url_to_host(url)
        assert_that(ret, equal_to('10.0.0.1'))

    def test_url_to_host_port(self):
        url = '10.0.0.1:90'
        ret = converter.url_to_host(url)
        assert_that(ret, equal_to('10.0.0.1'))

    def test_url_to_host_suffix(self):
        cidr = '10.0.0.1/32'
        address, prefix, mask = converter.parse_host_address(cidr)
        assert_that(address, equal_to('10.0.0.1'))
        assert_that(prefix, equal_to(32))

    def test_parse_host_address_illegal_multi_prefix(self):
        cidr = '10.0.0.1/32/16'

        def expect_error():
            converter.parse_host_address(cidr)
        assert_that(expect_error, raises(ValueError))

    def test_parse_host_address_no_ip(self):
        cidr = 'abc.com/32'
        address, prefix, mask = converter.parse_host_address(cidr)
        assert_that(address, 'abc.com')
        assert_that(prefix, none())
        assert_that(mask, none())

    def test_parse_host_address_normal(self):
        cidr = '10.244.211.30/24'
        address, prefix, mask = converter.parse_host_address(cidr)
        assert_that(address, equal_to('10.244.211.30'))
        assert_that(mask, equal_to('255.255.255.0'))
        assert_that(prefix, equal_to(24))

    def test_parse_host_address_29_bit(self):
        cidr = '10.244.211.30/30'
        address, prefix, mask = converter.parse_host_address(cidr)
        assert_that(address, equal_to('10.244.211.30'))
        assert_that(mask, equal_to('255.255.255.252'))
        assert_that(prefix, equal_to(30))

    def test_parse_host_address_7_bit(self):
        cidr = '10.244.211.30/7'
        address, prefix, mask = converter.parse_host_address(cidr)
        assert_that(address, equal_to('10.244.211.30'))
        assert_that(mask, equal_to('254.0.0.0'))
        assert_that(prefix, equal_to(7))

    def test_parse_host_address_illegal(self):
        cidr = '10.244.211.30/router'

        def expect_error():
            return converter.parse_host_address(cidr)
        assert_that(expect_error, raises(ValueError))

    def test_parse_host_address_illegal_upper(self):
        cidr = '10.244.211.30/33'

        def expect_error():
            return converter.parse_host_address(cidr)

        assert_that(expect_error, raises(ValueError))

    def test_parse_host_address_illegal_under(self):
        cidr = '10.244.211.30/-1'

        def expect_error():
            return converter.parse_host_address(cidr)

        assert_that(expect_error, raises(ValueError))

    def test_parse_host_address_ipv6_cidr(self):
        cidr = '2001:db8:a0b:12f0::/64'
        address, prefix, mask = converter.parse_host_address(cidr)
        assert_that(address, equal_to('2001:db8:a0b:12f0::'))
        assert_that(mask, equal_to('ffff:ffff:ffff:ffff:0000:0000:0000:0000'))
        assert_that(prefix, equal_to(64))

    def test_parse_host_address_ipv6_cidr2(self):
        cidr = '[2001:db8:a0b:12f0::/64]'
        address, prefix, mask = converter.parse_host_address(cidr)
        assert_that(address, equal_to('2001:db8:a0b:12f0::'))
        assert_that(mask,
                    equal_to('ffff:ffff:ffff:ffff:0000:0000:0000:0000'))
        assert_that(prefix, equal_to(64))

    def test_url_to_mask_ipv6_prefix_63(self):
        cidr = '2001:db8:a0b:12f0::/63'
        address, prefix, mask = converter.parse_host_address(cidr)
        assert_that(address, equal_to('2001:db8:a0b:12f0::'))
        assert_that(mask, equal_to('ffff:ffff:ffff:fffe:0000:0000:0000:0000'))
        assert_that(prefix, equal_to(63))

    def test_url_to_mask_ipv6_prefix_65(self):
        cidr = '2001:db8:a0b:12f0::/65'
        address, prefix, mask = converter.parse_host_address(cidr)
        assert_that(address, equal_to('2001:db8:a0b:12f0::'))
        assert_that(mask, equal_to('ffff:ffff:ffff:ffff:8000:0000:0000:0000'))
        assert_that(prefix, equal_to(65))

    def test_url_to_mask_ipv6_illegal_upper(self):
        cidr = '2001:db8:a0b:12f0::/129'

        def expect_error():
            converter.parse_host_address(cidr)

        assert_that(expect_error, raises(ValueError))

    def test_url_to_mask_ipv6_illegal_under(self):
        cidr = '2001:db8:a0b:12f0::/-1'

        def expect_error():
            converter.parse_host_address(cidr)

        assert_that(expect_error, raises(ValueError))

    def test_to_minute(self):
        r = converter.to_minute("00:06:00.000")
        assert_that(r, equal_to(6))

    def test_to_minute_invalid(self):
        r = converter.to_minute("07:00:00.000")
        assert_that(r, none())

    def test_to_hour(self):
        r = converter.to_hour("07:00:00.000")
        assert_that(r, equal_to(7))

    def test_from_minute(self):
        r = converter.from_minute(8)
        assert_that(r, equal_to("00:08:00.000"))

    def test_from_minite_str_lt_60(self):
        def conv():
            converter.from_minute(70)

        assert_that(conv, raises(ValueError))

    def test_from_hour_str(self):
        r = converter.from_hour(11)
        assert_that(r, equal_to("11:00:00.000"))

    def test_from_hour_str_lt_24(self):
        def conv():
            converter.from_hour(25)

        assert_that(conv, raises(ValueError))

    def test_to_raid_type_list(self):
        r = converter.to_raid_type_list('')
        assert_that(len(r), equal_to(0))
        r = converter.to_raid_type_list('r1 r1_0 hot_spare Unbound')
        assert_that(r[0], equal_to(VNXRaidType.RAID1))
        assert_that(r[1], equal_to(VNXRaidType.RAID10))
        assert_that(r[2], equal_to(VNXRaidType.HOTSPARE))
        assert_that(r[3], equal_to(VNXRaidType.UNBOUND))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import sys
import logging

from storops.lib.thinclone_helper import TCHelper  # noqa
from storops.unity.enums import *  # noqa
from storops.unity.resource.system import UnitySystem  # noqa
from storops.unity.resource.snap_schedule import UnitySnapScheduleRule  # noqa
from storops.vnx.enums import *  # noqa
from storops.vnx.resource.system import VNXSystem  # noqa

__author__ = 'Cedric Zhuang'


def enable_log(level=logging.DEBUG):
    """Enable console logging.

    This is a utils method for try run with storops.
    :param level: log level, default to DEBUG
    """
    logger = logging.getLogger(__name__)
    logger.setLevel(level)
    if not logger.handlers:
        logger.info('enabling logging to console.')
        logger.addHandler(logging.StreamHandler(sys.stdout))


def disable_log():
    logger = logging.getLogger(__name__)
    logger.info('disabling logging to console.')
    logger.setLevel(logging.NOTSET)
    logger.handlers = []
    return logger



# coding=utf-8
# Copyright (c) 2017 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, equal_to, instance_of, raises, is_not

from storops_test.vnx.cli_mock import t_cli, patch_cli

from storops.vnx.resource.nqm import VNXIOClass, VNXIOClassList, VNXIOPolicy, \
    VNXLun, normalize_lun

__author__ = "Peter Wang"


class VNXIOClassTest(TestCase):
    def test_normalize_lun(self):
        def _inner():
            normalize_lun('invalid', t_cli())
        assert_that(_inner, raises(ValueError, 'Invalid format'))

    @patch_cli
    def test_ioclass_all(self):
        ioclass_list = VNXIOClassList(cli=t_cli())
        assert_that(len(ioclass_list), equal_to(6))
        assert_that(ioclass_list[0], instance_of(VNXIOClass))

    @patch_cli
    def test_ioclass(self):
        ioclass = VNXIOClass(name='with_luns_snaps', cli=t_cli())
        assert_that(ioclass.name, equal_to('with_luns_snaps'))
        assert_that(ioclass.state, equal_to('Idle'))
        assert_that(ioclass.io_type, equal_to('ReadWrite'))
        assert_that(ioclass.io_size_range, equal_to('Any'))
        assert_that(ioclass.control_method, equal_to('Limit'))
        assert_that(ioclass.metric_type, equal_to('Bandwidth'))
        assert_that(ioclass.goal_value, equal_to('1000.0 MB/s'))
        map(lambda l: assert_that(l, instance_of(VNXLun)),
            ioclass.luns)
        assert_that(len(ioclass.luns), equal_to(2))
        assert_that(ioclass.policy, instance_of(VNXIOPolicy))
        assert_that(ioclass.policy.name, equal_to('new_name_test'))

    @patch_cli
    def test_class_multiple_luns(self):
        ioclass = VNXIOClass(name='with_3_luns', cli=t_cli())
        assert_that(ioclass.state, equal_to('Idle'))
        assert_that(ioclass.io_type, equal_to('ReadWrite'))
        assert_that(ioclass.io_size_range, equal_to('Any'))
        assert_that(ioclass.control_method, equal_to('No Control'))
        map(lambda l: assert_that(l, instance_of(VNXLun)),
            ioclass.luns)
        assert_that(len(ioclass.luns), equal_to(3))

    @patch_cli
    def test_ioclass_nolun(self):
        ioclass = VNXIOClass(name='simple', cli=t_cli())
        assert_that(ioclass.name, equal_to('simple'))
        assert_that(ioclass.state, equal_to('Idle'))
        assert_that(ioclass.status, equal_to('Ready'))
        assert_that(len(ioclass.luns), equal_to(0))

    @patch_cli
    def test_create_ioclass(self):
        ioclass = VNXIOClass.create(cli=t_cli(), name="simple",
                                    iotype='rw')
        assert_that(ioclass.name, equal_to('simple'))

    @patch_cli
    def test_create_ioclass_with_lun(self):
        lun2 = VNXLun(name='lun2', cli=t_cli())
        ioclass = VNXIOClass.create(cli=t_cli(), name="simple",
                                    iotype='rw', luns=lun2)
        assert_that(ioclass.name, equal_to('simple'))

    @patch_cli
    def test_delete_ioclasse(self):
        ioclass = VNXIOClass(name='to_delete', cli=t_cli())
        deleted = ioclass.delete()
        assert_that(deleted.existed, equal_to(False))

    @patch_cli
    def test_add_lun_to_ioclass(self):
        ioclass = VNXIOClass(name='with_luns_snaps', cli=t_cli())
        lun1 = VNXLun(lun_id=1, cli=t_cli())
        ioclass.add_lun(lun1)
        # Add via lun id
        ioclass.add_lun(2)

    @patch_cli
    def test_add_smp_to_ioclass(self):
        ioclass = VNXIOClass(name='with_luns_snaps', cli=t_cli())
        smp1 = VNXLun(name='m2', cli=t_cli())
        new_class = ioclass.add_lun(smp1)
        assert_that(new_class, instance_of(VNXIOClass))
        assert_that(new_class, is_not(ioclass))

    @patch_cli
    def test_add_lun_to_running_ioclass(self):
        ioclass = VNXIOClass(name='running_ioclass', cli=t_cli())
        lun1 = VNXLun(lun_id=1, cli=t_cli())

        ioclass.add_lun(lun1)

    @patch_cli
    def test_remove_lun(self):
        ioclass = VNXIOClass(name='with_luns_snaps', cli=t_cli())
        ioclass.remove_lun(0)
        lun1 = VNXLun(lun_id=2, cli=t_cli())
        ioclass.remove_lun(lun1)

    @patch_cli
    def test_add_to_policy(self):
        ioclass = VNXIOClass(name='simple', cli=t_cli())
        policy = VNXIOPolicy(name='with_ioclass', cli=t_cli())
        ioclass.add_to_policy(policy)


class VNXIOPolicyTest(TestCase):
    @staticmethod
    def get_policy():
        return VNXIOPolicy(name='simple')

    @patch_cli
    def test_policy(self):
        policy = VNXIOPolicy(name='with_ioclass', cli=t_cli())
        assert_that(policy.name, equal_to('with_ioclass'))
        assert_that(policy.status, equal_to('Ready'))
        assert_that(policy.state, equal_to('Idle'))

    @patch_cli
    def test_create_policy(self):
        policy = VNXIOPolicy.create(name='new_policy', cli=t_cli())
        assert_that(policy.status, equal_to('Warning'))

    @patch_cli
    def test_delete_policy(self):
        policy = VNXIOPolicy(name='to_delete', cli=t_cli())
        deleted = policy.delete()
        assert_that(deleted.existed, equal_to(False))

    @patch_cli
    def test_stop_policy(self):
        VNXIOPolicy.stop_policy(cli=t_cli())

    @patch_cli
    def test_run_policy(self):
        policy = VNXIOPolicy(name='simple', cli=t_cli())
        policy.run_policy()

    @patch_cli
    def test_measure_policy(self):
        policy = VNXIOPolicy(name='simple', cli=t_cli())
        policy.measure_policy()

    @patch_cli
    def test_add_class_to_policy(self):
        policy = VNXIOPolicy(name='with_ioclass', cli=t_cli())
        ioclass = VNXIOClass(name='with_luns_snaps', cli=t_cli())
        policy.add_class(ioclass)

    @patch_cli
    def test_add_class_to_running_policy(self):
        policy = VNXIOPolicy(name='running_policy', cli=t_cli())
        ioclass = VNXIOClass(name='with_luns_snaps', cli=t_cli())

        policy.add_class(ioclass)

    @patch_cli
    def test_remove_class_already_removed(self):
        policy = VNXIOPolicy(name='with_ioclass', cli=t_cli())
        ioclass = VNXIOClass(name='with_luns_snaps', cli=t_cli())
        policy.remove_class(ioclass)

    @patch_cli
    def test_remove_class_from_existing(self):
        policy = VNXIOPolicy(name='with_ioclass', cli=t_cli())
        ioclass = VNXIOClass(name='simple', cli=t_cli())
        policy.remove_class(ioclass)



# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import logging

from storops.unity.resource import UnityResource, UnityResourceList
from storops.lib.version import version
import storops.unity.resource.nas_server

__author__ = 'Tina Tang'

LOG = logging.getLogger(__name__)


@version('>=4.1')
class UnityTenant(UnityResource):
    @classmethod
    def create(cls, cli, name, uuid=None, vlans=None):
        resp = cli.post(cls().resource_class,
                        name=name,
                        uuid=uuid,
                        vlans=vlans)
        resp.raise_if_err()
        return cls(_id=resp.resource_id, cli=cli)

    def modify(self, name=None, vlans=None):
        req_body = {}
        if name is not None:
            req_body['name'] = name
        if vlans is not None:
            req_body['vlans'] = vlans
        resp = self._cli.modify(self.resource_class,
                                self.get_id(), **req_body)
        resp.raise_if_err()
        return resp

    def delete(self, delete_hosts=False):
        if delete_hosts and self.hosts:
            for host in self.hosts:
                host.delete()
        return super(UnityTenant, self).delete()

    @property
    def nas_servers(self):
        clz = storops.unity.resource.nas_server.UnityNasServerList
        return clz(cli=self._cli, tenant=self)


@version('>=4.1')
class UnityTenantList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityTenant



# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import logging

from storops.unity.resource import UnityResource, UnityResourceList

LOG = logging.getLogger(__name__)


class UnityAlert(UnityResource):
    def modify(self, is_acknowledged=None):
        req_body = self._cli.make_body(isAcknowledged=is_acknowledged)
        resp = self.action('modify', **req_body)
        resp.raise_if_err()
        return resp

    def delete(self):
        resp = self._cli.delete(self.resource_class,
                                self.get_id())
        resp.raise_if_err()
        return resp


class UnityAlertList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityAlert



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from storops.vnx.resource import VNXCliResource, VNXCliResourceList
from storops import exception as ex
from storops.lib.common import round_3
from storops.lib.converter import block_to_gb

__author__ = 'Cedric Zhuang'


class VNXRaidGroup(VNXCliResource):
    def __init__(self, raid_group_id=None, cli=None):
        super(VNXRaidGroup, self).__init__()
        self._cli = cli
        self._raid_group_id = raid_group_id

    def _get_raid_group_id(self):
        if self._raid_group_id is not None:
            ret = self._raid_group_id
        else:
            ret = self.raid_group_id
        return ret

    def _get_raw_resource(self):
        return self._cli.get_rg(rg_id=self._raid_group_id, poll=self.poll)

    @staticmethod
    def create(cli, raid_group_id, disks, raid_type=None):
        ret = cli.create_rg(disks, raid_group_id, raid_type)
        ex.raise_if_err(ret, default=ex.VNXCreateRaidGroupError)
        return VNXRaidGroup(raid_group_id, cli)

    def delete(self):
        ret = self._cli.delete_rg(self._get_raid_group_id(), poll=self.poll)
        ex.raise_if_err(ret, default=ex.VNXDeleteRaidGroupError)

    @classmethod
    def get(cls, cli, raid_group_id=None):
        if raid_group_id is None:
            ret = VNXRaidGroupList(cli)
        else:
            ret = VNXRaidGroup(raid_group_id, cli)
        return ret

    @property
    @round_3
    def available_capacity_gbs(self):
        return block_to_gb(self.free_capacity_blocks_non_contiguous)


class VNXRaidGroupList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXRaidGroup

    def _get_raw_resource(self):
        return self._cli.get_rg(poll=self.poll)
